#include "stdafx.h"
#include "Tony.h"
#include "InverseLinear.h"
#include "InverseBinary.h"
#include "Alias.h"
#include "RandomCRT.h"
#include "RandomLCG.h"

using namespace std;

///////////////////////////////////////////////////////////////////////////////
// Profiling

// For storing results of profiles, and generating google API chart
struct Profile {
	Profile(const std::string& name, size_t N, double init, double sample, double error, size_t memory) : 
		name(name), N(N), init(init), sample(sample), error(error), memory(memory)
	{
	}

	bool operator<(const Profile& rhs) const {
		if (N == rhs.N)
			return name < rhs.name;
		else
			return N < rhs.N;
	}

	std::string name;
	size_t N;
	double init;
	double sample;
	double error;
	size_t memory;
};

std::vector<Profile> gProfiles;

struct SampleAccessor { static double Get(const Profile& p) { return p.sample; } };
struct InitAccessor { static double Get(const Profile& p) { return p.init; } };
struct ErrorAccessor { static double Get(const Profile& p) { return p.error; } };
struct MemoryAccessor { static double Get(const Profile& p) { return p.memory; } };

void WriteTable(const char* filename) {
	ofstream os(filename);
	
	os << fixed; // don't use scientific
	os << "<html>" << endl;
	os << "  <head>" << endl;
	os << "    <script type='text/javascript' src='http://www.google.com/jsapi'></script>" << endl;
	os << "    <script type='text/javascript'>" << endl;
	os << "      google.load('visualization', '1', {packages:['table']});" << endl;
	os << "      google.setOnLoadCallback(drawTable);" << endl;
	os << "      function drawTable() {" << endl;
	os << "        var data = new google.visualization.DataTable();" << endl;
	os << "        data.addColumn('string', 'sampler');" << endl;
	os << "        data.addColumn('number', 'N');" << endl;
	os << "        data.addColumn('number', 'init');" << endl;
	os << "        data.addColumn('number', 'sample');" << endl;
	os << "        data.addColumn('number', 'error');" << endl;
	os << "        data.addColumn('number', 'memory');" << endl;
	
	for (vector<Profile>::iterator i = gProfiles.begin(); i != gProfiles.end(); ++i)
		if (i->sample > 0 && i->memory > 0)
			os << "        data.addRow(['" << i->name << "'," << i->N << "," << i->init << "," << i->sample << "," << i->error << "," << i->memory << "]);" << endl;

	os << "	var table = new google.visualization.Table(document.getElementById('table_div'));" << endl;
	os << "	table.draw(data, {width: 500});" << endl;
	os << "}" << endl;
	os << "</script>" << endl;
	os << "</head>" << endl;
	os << "<body>" << endl;
	os << "<div id='table_div'></div>" << endl;
	os << "</body>" << endl;
	os << "</html>" << endl;
}

template <typename FieldAccessor>
void WriteChart(const char* filename, const char* title, const char* options = 0) {
	ofstream os(filename);

	os << fixed; // don't use scientific
	os << "<html>" << endl;
	os << "  <head>" << endl;
	os << "    <script type='text/javascript' src='http://www.google.com/jsapi'></script>" << endl;
	os << "    <script type='text/javascript'>" << endl;
	os << "      google.load('visualization', '1', {packages:['corechart']});" << endl;
	os << "      google.setOnLoadCallback(drawChart);" << endl;
	os << "      function drawChart() {" << endl;
	os << "        var data = new google.visualization.DataTable();" << endl;
	os << "        data.addColumn('string', 'N');" << endl;

	size_t caseCount = 0;
	for (vector<Profile>::iterator i = gProfiles.begin(); i->N == gProfiles.front().N; ++i) {
		os << "        data.addColumn('number', '" << i->name <<"');" << endl;
		caseCount++;
	}

	assert(gProfiles.size() % caseCount == 0);
	for (vector<Profile>::iterator i = gProfiles.begin(); i != gProfiles.end(); ) {
		size_t N = i->N;
		os << "        data.addRow(['" << N << "',";
		for (size_t c = 0; c < caseCount; ++c, ++i) {
			assert(i->N == N);
			if (i->sample > 0)
				os << FieldAccessor::Get(*i);

			if (c != caseCount - 1)
				os << ",";
		}
		os << "]);" << endl;
	}
	os << "        var chart = new google.visualization.LineChart(document.getElementById('chart_div'));" << endl;
	os << "        chart.draw(data, {width: 800, height: 800, title: '" << title << "'";
	if (options)
		os << "," << options;
	os << "});" << endl;
	os << "      }" << endl;
	os << "    </script>" << endl;
	os << "  </head>" << endl;
	os << endl;
	os << "  <body>" << endl;
	os << "    <div id='chart_div'></div>" << endl;
	os << "  </body>" << endl;
	os << "</html>" << endl;
}	

template <typename T>
T ComputeError(const unsigned* histogram, const T* pdf, size_t N, size_t sampleCount) {
	T error = 0.0;
	for (size_t i = 0; i < N; i++) {
		T delta = (T)histogram[i] / sampleCount - pdf[i];
		error += abs(delta);
	}
	return error;
}

double ComputeDuration(LARGE_INTEGER start, LARGE_INTEGER end) {
	LARGE_INTEGER freq;
	QueryPerformanceFrequency(&freq);
	return double(end.QuadPart - start.QuadPart) / freq.QuadPart;
}

///////////////////////////////////////////////////////////////////////////////
// Utilities

template<typename T>
inline void Pdf2Cdf(const T* pdf, T* cdf, size_t N) {
	assert(N > 0);

	cdf[0] = pdf[0];

    for (size_t i = 1; i < N - 1; i++)
        cdf[i] = cdf[i - 1] + pdf[i];

    // Force set last cdf to 1, preventing floating-point summing error in the loop.
    cdf[N - 1] = 1;
}

void Pdf2PerctangePdf(const double* pdf, double* percentagePdf, size_t N) {
	for (size_t i = 0; i < N; i++)
		percentagePdf[i] = pdf[i] * 100.0;
}

void RandomizePdf(size_t N, vector<double>& pdf, vector<float>& floatpdf) {
	pdf.resize(N);
	floatpdf.resize(N);

	// generate random pdf
	double sum = 0.0;
	for (size_t i = 0; i < N; i++) {
		pdf[i] = rand() * rand();
		sum += pdf[i];
	}

	// Normalize
	for (size_t i = 0; i < N; i++) {
		pdf[i] /= sum;
		floatpdf[i] = (float)pdf[i];
	}
}

///////////////////////////////////////////////////////////////////////////////
// Experiment 1

void TestTony(const double* percentagePdf, const double*pdf, size_t N, size_t initCount, size_t sampleCount) {
	if (N > 100) {
		gProfiles.push_back(Profile("Tony", N, 0, 0, 0, 0));
		return;
	}

	// initialization timing
	LARGE_INTEGER initStart, initEnd;
	QueryPerformanceCounter(&initStart);
	for (size_t i = 0; i < initCount; i++)
		CTonyRandomArea Area1(const_cast<double*>(percentagePdf), N);
	QueryPerformanceCounter(&initEnd);

	// sampling timing
	vector<unsigned> histogram(N);

	srand(0);	// Reset seed for deterministic result

    CTonyRandomArea Area1(const_cast<double*>(percentagePdf), N);

	LARGE_INTEGER start, end;
	QueryPerformanceCounter(&start);
	for (size_t i = 0; i < sampleCount; i++) {
		size_t sample = Area1.GetRandomType();
		assert(sample >= 0);
		assert(sample < N);
		histogram[sample]++;
	}
	QueryPerformanceCounter(&end);

	cout << setw(20) << "Tony"
		<< " N=" << setw(3) << N 
		<< " Init=" << ComputeDuration(initStart, initEnd)
		<< " Err=" << ComputeError(&histogram[0], pdf, N, sampleCount) 
		<< " Sample=" << ComputeDuration(start, end)
		<< " Mem=" << Area1.GetMemory()
		<< endl;

	gProfiles.push_back(Profile("Tony", N, ComputeDuration(initStart, initEnd), ComputeDuration(start, end), ComputeError(&histogram[0], pdf, N, sampleCount), Area1.GetMemory()));
}

template <typename T, typename Sampler>
void TestSingle(Sampler& sampler, const T* pdf, size_t N, size_t sampleCount) {
	vector<unsigned> histogram(N);

	LARGE_INTEGER start, end;
	QueryPerformanceCounter(&start);
	for (size_t i = 0; i < sampleCount; i++)
	{
		size_t sample = sampler();
		assert(sample >= 0);
		assert(sample < N);
		histogram[sample]++;
	}
	QueryPerformanceCounter(&end);
	
	cout << " Err=" << ComputeError(&histogram[0], pdf, N, sampleCount)
		<< " Sample=" << ComputeDuration(start, end)
		<< " Mem=" << sampler.GetMemory() << endl;

	// Assume the profile of this test is already push_back-ed
	gProfiles.back().sample = ComputeDuration(start, end);
	gProfiles.back().error = ComputeError(&histogram[0], pdf, N, sampleCount);
	gProfiles.back().memory = sampler.GetMemory();
}

template <typename T,typename RandomGenerator>
void TestMultiple(const T* pdf, size_t N, size_t initCount, size_t sampleCount, RandomGenerator random, const char* caseName) {
	vector<T> cdf(N);
	Pdf2Cdf(pdf, &cdf[0], N);

	std::string s(caseName);
	LARGE_INTEGER initStart, initEnd;

	// InverseLinear
	QueryPerformanceCounter(&initStart);
	for (size_t i = 0; i < initCount; i++)
		InverseLinear<T, RandomGenerator> sampler(&cdf[0], N, random);
	QueryPerformanceCounter(&initEnd);
	
	cout << setw(20) << "Linear<" + s + ">"
		<< " N=" << setw(3) << N
		<< " Init=" << ComputeDuration(initStart, initEnd);
	gProfiles.push_back(Profile("Linear<" + s + ">", N, ComputeDuration(initStart, initEnd), 0, 0, 0));
	TestSingle<T>(InverseLinear<T, RandomGenerator>(&cdf[0], N, random), pdf, N, sampleCount);

	// InverseBinary
	QueryPerformanceCounter(&initStart);
	for (size_t i = 0; i < initCount; i++)
		InverseBinary<T, RandomGenerator> sampler(&cdf[0], N, random);
	QueryPerformanceCounter(&initEnd);

	cout << setw(20) << "Binary<" + s + ">"
		<< " N=" << setw(3) << N
		<< " Init=" << ComputeDuration(initStart, initEnd);
	gProfiles.push_back(Profile("Binary<" + s + ">", N, ComputeDuration(initStart, initEnd), 0, 0, 0));
	TestSingle<T>(InverseBinary<T, RandomGenerator>(&cdf[0], N, random), pdf, N, sampleCount);

	// Alias
	QueryPerformanceCounter(&initStart);
	for (size_t i = 0; i < initCount; i++)
		Alias<T, RandomGenerator> sampler(&pdf[0], N, random);
	QueryPerformanceCounter(&initEnd);

	cout << setw(20) << "Alias<" + s + ">"
		<< " N=" << setw(3) << N
		<< " Init=" << ComputeDuration(initStart, initEnd);
	gProfiles.push_back(Profile("Alias<" + s + ">", N, ComputeDuration(initStart, initEnd), 0, 0, 0));
	TestSingle<T>(Alias<T, RandomGenerator>(&pdf[0], N, random), pdf, N, sampleCount);
}

void TestPredefinedPDF(const size_t initCount, const size_t sampleCount)
{
	static const size_t N = 5;
	static const double pdf[N] = { 0.12, 0.40, 0.40, 0.07, 0.01 };
	static const double percentagePdf[N] = { 12.0, 40.0, 40.0, 7.0, 1.0 };

	// Convert pdf to float
	float floatPdf[N];
	for (size_t i = 0; i < N; i++)
		floatPdf[i] = (float)pdf[i];

	cout << "Predefined PDF with N=" << N << endl;

	TestTony(percentagePdf, pdf, N, initCount, sampleCount);
	TestMultiple<double>(pdf, N, initCount, sampleCount, RandomCRT<double>(), "CRT, double");
	TestMultiple<double>(pdf, N, initCount, sampleCount, RandomLCG<double>(), "LCG, double");
	TestMultiple<float>(floatPdf, N, initCount, sampleCount, RandomLCG<float>(), "LCG, float");

	cout << endl;
}

void TestTonyCrashCase( const size_t initCount, const size_t sampleCount )
{
	static const size_t N = 5;
	static const double pdf[N] = { 0.0, 0.0, 0.0, 0.0, 0.0 };	// as placeholder, not actually used here
	static const double percentagePdf[N] = { 3, 5, 7, 11, 13 };
	TestTony(percentagePdf, pdf, N, initCount, sampleCount);
}

// Testing random pdf table of size N = {4, 8, ... }
void TestRandomPdfVariousN( const size_t initCount, const size_t sampleCount )
{
	// Not need to store the previous profiling results
	gProfiles.clear();

	cout << "Randomly generated PDF with various N" << endl;
	size_t N = 4;
	vector<double> pdf;
	vector<float> floatpdf; 
	vector<double> percentagePdf(N);

	for (int iteration = 0; iteration < 7; iteration++, N *= 2) {
		RandomizePdf(N, pdf, floatpdf);

		percentagePdf.resize(N);
		Pdf2PerctangePdf(&pdf[0], &percentagePdf[0], N);

		TestTony(&percentagePdf[0], &pdf[0],  N, initCount, sampleCount);
		TestMultiple<double>(&pdf[0], N, initCount, sampleCount, RandomCRT<double>(), "CRT, double");
		TestMultiple<double>(&pdf[0], N, initCount, sampleCount, RandomLCG<double>(), "LCG, double");
		TestMultiple<float>(&floatpdf[0], N, initCount, sampleCount, RandomLCG<float>(), "LCG, float");
	}

	// Write google charts
	WriteTable("table.html");
	WriteChart<SampleAccessor>("sample.html", "Sampling Time");
	WriteChart<InitAccessor>("init.html", "Initialization Time");
	WriteChart<ErrorAccessor>("error.html", "Error");
	WriteChart<MemoryAccessor>("memory.html", "Memory");
}

void TestTonyInterleaved(const size_t M, const size_t N, vector<double>* pdfs, const size_t interleavedSample)
{
	vector<double> percentagePdf(N);
	vector<CTonyRandomArea*> samplers(M);
	size_t memory = 0;

	LARGE_INTEGER initStart, initEnd;
	QueryPerformanceCounter(&initStart);
	for (size_t j = 0; j < M; j++) {
		Pdf2PerctangePdf(&pdfs[j][0], &percentagePdf[0], N);
		samplers[j] = new CTonyRandomArea(const_cast<double*>(&percentagePdf[0]), N);
		memory += samplers[j]->GetMemory();
	}
	QueryPerformanceCounter(&initEnd);

	// Shuffle order of samplers
	random_shuffle(samplers.begin(), samplers.end());

	size_t unused = 0; // prevent optimized out
	LARGE_INTEGER start, end;
	QueryPerformanceCounter(&start);
	for (size_t i = 0; i < interleavedSample; i++)
		for (size_t j = 0; j < M; j++)
			unused += samplers[j]->GetRandomType();
	QueryPerformanceCounter(&end);
	cout << unused << endl;

	for (size_t j = 0; j < M; j++)
		delete samplers[j];

	gProfiles.push_back(Profile("Tony", M, ComputeDuration(initStart, initEnd), ComputeDuration(start, end), 0, memory));
}

///////////////////////////////////////////////////////////////////////////////
// Experiment 2

template <typename Sampler>
void TestSingleInterleaved(const size_t M, const size_t N, vector<float>* floatpdfs, const size_t interleavedSample, const char* caseName, bool samplerUseCdf)
{
	vector<Sampler*> samplers(M);
	vector<float> cdf(N * M);	// use a continuous vector to store all cdfs
	size_t memory = 0;

	LARGE_INTEGER initStart, initEnd;
	QueryPerformanceCounter(&initStart);
	for (size_t j = 0; j < M; j++) {
		if (Sampler::AcceptCdf) {
			Pdf2Cdf(&floatpdfs[j][0], &cdf[j * N], N);
			samplers[j] = new Sampler(&cdf[j * N], N, Sampler::RandomGeneratorType());
		}
		else
			samplers[j] = new Sampler(&floatpdfs[j][0], N, Sampler::RandomGeneratorType());

		memory += samplers[j]->GetMemory();
	}
	QueryPerformanceCounter(&initEnd);

	// Shuffle order of samplers
	random_shuffle(samplers.begin(), samplers.end());

	size_t unused = 0; // prevent optimized out
	LARGE_INTEGER start, end;
	QueryPerformanceCounter(&start);
	for (size_t i = 0; i < interleavedSample; i++)
		for (size_t j = 0; j < M; j++)
			unused += (*samplers[j])();
	QueryPerformanceCounter(&end);
	cout << unused << endl;
 
	for (size_t j = 0; j < M; j++)
		delete samplers[j];

	gProfiles.push_back(Profile(caseName, M, ComputeDuration(initStart, initEnd), ComputeDuration(start, end), 0, memory));
}

// Testing interleaved multiple samplers
void TestInterleaved(const size_t sampleCount)
{
	gProfiles.clear();

	const size_t N = 100;

	cout << "Randomly generated PDF with N=" << N << " and various M interleaved samplers" << endl;

	size_t M = 1;
	for (int iteration = 0; iteration < 6; iteration++, M *= 10) {
		assert(sampleCount % M == 0);
		const size_t interleavedSample = sampleCount / M;
		vector<double>* pdfs = new vector<double>[M];
		vector<float>* floatpdfs = new vector<float>[M]; 

		for (size_t j = 0; j < M; j++)
			RandomizePdf(N, pdfs[j], floatpdfs[j]);

		TestTonyInterleaved(M, N, pdfs, interleavedSample);
		TestSingleInterleaved<InverseBinary<float, RandomLCG<float>>>(M, N, floatpdfs, interleavedSample, "Binary<LCG, float>", true);
		TestSingleInterleaved<Alias<float, RandomLCG<float>>>(M, N, floatpdfs, interleavedSample, "Alias<LCG, float>", false);

		delete [] pdfs;
		delete [] floatpdfs;
	}

	WriteTable("interleavedTable.html");
	WriteChart<SampleAccessor>("interleavedSample.html", "Sampling time with N=100, M samplers");
}

///////////////////////////////////////////////////////////////////////////////
// Main

int main(int argc, char* argv[]) {
	static const size_t initCount = 1000;
	static const size_t sampleCount = 1000000;

	cout << fixed; // don't use scientific
	cout << "InitCount=" << initCount << " SampleCount=" << sampleCount << endl << endl;

	TestPredefinedPDF(initCount, sampleCount);
	TestTonyCrashCase(initCount, sampleCount);
	TestRandomPdfVariousN(initCount, sampleCount);
	TestInterleaved(sampleCount);

	return 0;
}
